#!/usr/bin/env python
import sys
import json
import socket
import time
import xml.etree.ElementTree as ET
from tetrationio.tetration import Tetration
from utils.errors import InvalidScopeError, AccessForbiddenError, QuarantineAnnotationNotAddedError, QuarantineAnnotationNotDeletedError, \
     AuthorizationError, CannotAccessTetrationError, AccessFailed, UnknownError, ErrorCode 

'''
@author Feng Liu (feliu@cisco.com)
Copyright (c) 2018 by Cisco Systems, Inc.

FMC remediation API:
(1) Module input is from 'instance.conf'
(2) Event data (src_ip here) is passed in through sys.argv[2]
The 'instance.conf' example content is as follows:
            <instance name="MyPyReme">
                <config>
                    <ipaddress name="tetration_addr">10.193.227.1</ipaddress>
                    <string  name="scope">Default</string>
                    <password name="api_key">***</password>
                    <password name="api_secret">***</password>
               </config>
               <remediation name="add-reme" type="quarantine">
                   <config>
                   </config>
               </remediation>
               <remediation name="delete-reme" type="un-quarantine">
                   <config>
                   </config>
               </remediation>
             </instance>

'''

logfile = "debug.log"
config = {}
remediation = {}
tree = ET.parse('instance.conf')
root = tree.getroot()
total_elements = len(root)
# element root[0] is the config, the rest is/are remediation element(s)
# parse config first
for child in root[0]:
    config[child.get('name')] = child.text
hosts = []
addr = config['tetration_addr']
if addr and addr.strip() != '':
    hosts.append(addr)

# now parse remediation element(s). For above example, we're getting a dictionary like this: {'add-reme': 'quarantine'; 'delete-reme': 'un-quarantine'}
for index in range (1, total_elements):
    remediation[root[index].get('name')] = root[index].get('type')

tetration = Tetration(hosts, config['scope'], config['api_key'], config['api_secret'])
remediation_name = sys.argv[1]
remediation_type = remediation[remediation_name]
action = 'add' if remediation_type == 'quarantine' else 'delete'
tetration.log_msg(logfile, "remediation_name=" + remediation_name, '')
tetration.log_msg(logfile, "remediation_type=" + remediation_type, '')
ip = sys.argv[2] # get the offending IP from FMC Remediation API call

start_at = time.time()
tetration.log_msg(logfile, "Start to handle " + ip + " at: ", time.ctime())

try:
    tetration.quarantine(ip, action)
except InvalidScopeError:
    sys.exit(ErrorCode.INVALID_SCOPE_ERROR)
except AccessForbiddenError:
    sys.exit(ErrorCode.ACCESS_FORBIDDEN_ERROR)
except QuarantineAnnotationNotAddedError:
    sys.exit(ErrorCode.QUARANTINE_ANNOTATION_NOT_ADDED_ERROR)
except QuarantineAnnotationNotDeletedError:
    sys.exit(ErrorCode.QUARANTINE_ANNONTATION_NOT_DELETED_ERROR)
except AuthorizationError:
    sys.exit(ErrorCode.AUTHORIZATION_ERROR)
except CannotAccessTetrationError:
    sys.exit(ErrorCode.CANNOT_ACCESS_TETRATION_ERROR)
except AccessFailed:
    sys.exit(ErrorCode.ACCESS_FAILED)
except UnknownError:
    sys.exit(ErrorCode.UNKNOWN_ERROR)
except Exception as e:
    tetration.log_msg(logfile, "Encountered error: " + str(e) + "at: ", time.ctime())
    sys.exit(ErrorCode.UNKNOWN_ERROR)
finally:
    end_at = time.time()
    tetration.log_msg(logfile, "Finished handling " + ip + " at: ", time.ctime())
    tetration.log_msg(logfile, "Time spent to handle " + ip + " is: ", str(end_at - start_at))

tetration.log_msg(logfile, "Finished remediation, program will exit with status 0 shortly at: ", time.ctime())
sys.exit(0)

