'''
Created on May 18, 2018

@author: feliu
Copyright (c) 2018 by Cisco Systems, Inc.
'''

import json
from  utils.errors import InvalidScopeError, QuarantineAnnotationNotAddedError, QuarantineAnnotationNotDeletedError, \
      AccessForbiddenError, AuthorizationError, CannotAccessTetrationError, AccessFailed, UnknownError
import time

from tetpyclient import RestClient

logfile = "debug.log"
debug = True

class Tetration(object):
    '''
    This represents the connection to an APIC target.
    '''

    def __init__(self, hosts, scope, api_key, api_secret):
        '''
        @param config: the name of the file that contains APIC access info. The config example content is as follows:
            <instance name="MyPyReme">
                <config>
                    <ipaddress name="tetration_addr">172.26.46.68</ipaddress>
                    <password name="api_key">***</string>
                    <password name="api_secret">***</password>
               </config>
               <remediation name="add-reme" type="quarantine">
                   <config>
                   </config>
               </remediation>
               <remediation name="delete-reme" type="un-quarantine">
                   <config>
                   </config>
               </remediation>
             </instance>
        @param tetration_addr: IP of the Tetration Analytics
        @param api_key: api_key to the Tetration Analytics
        @param api_secret: api_secret to the Tetration Analytics
        '''

        self.init()
        hosts.sort()
        self.hosts = hosts
        self.active_ip = hosts[0]
        self.scope = scope
        self.dic['api_key'] = api_key
        self.dic['api_secret'] = api_secret
        # use api_key/api_secret to re-create file 'api_credentials.json'
        self.log_msg('debug.log', 'active_ip=', self.active_ip)
        self.log_msg('debug.log', 'api_key=', api_key)
        self.log_msg('debug.log', 'api_secret=', api_secret)
        with open('api_credentials.json', 'w+') as f:
            f.write('{\n')
            f.write('  "api_key": "' + api_key + '",\n')
            f.write('  "api_secret": "' + api_secret + '"\n')
            f.write('}\n')

    def init(self):
        self.dic = {}

    def quarantine(self, ip, action):
        self.log_msg('debug.log', 'ip=', ip)
        self.log_msg('debug.log', 'scope=', self.scope)
        if self.scope is None or not self.scope.isalnum():
            # As scope is going to be used as part of url, we do input validation here again to make sure no special charators is allowed to avoid potential security risk.
            raise InvalidScopeError("Invalid scope. Scope must be alphabetic or numeric charactors only.")
        restclient = RestClient("https://" + self.active_ip, credentials_file="api_credentials.json", verify=False)
        try:
            if action == 'add':
                req_payload = {'ip': ip, 'attributes': {'quarantine': 'yes'}}
                resp = restclient.post('/inventory/tags/' + self.scope, json_body=json.dumps(req_payload))
            else:
                req_payload = {'ip': ip, 'attributes': {'quarantine': 'no'}}
                resp = restclient.post('/inventory/tags/' + self.scope, json_body=json.dumps(req_payload))
        except Exception as ex:
            self.log_msg('debug.log', 'exception=', str(ex))
            raise CannotAccessTetrationError("Cannot access Secure Workload. Check Secure Workload IP and connectivity.")
        if resp.status_code == 200:
            parsed_resp = json.loads(resp.content) if resp.content not in ['', b'']  else ''
            self.log_msg('debug.log', 'parsed_resp=', str(parsed_resp))
            print(json.dumps(parsed_resp, indent=4, sort_keys=True))
            if action == 'add' and not self.is_quarantine_annotation_present(ip, action):
                raise QuarantineAnnotationNotAddedError("The quarantine annotation was not added, check the connection to Secure Workload and if the Scope is correct.")
            elif action == 'delete' and self.is_quarantine_annotation_present(ip, action):
                raise QuarantineAnnotationNotDeletedError("The quarantine annotation was not deleted, check the connection to Secure Workload and if the Scope is correct.")
        elif resp.status_code == 400 and 'invalid scope name in request URL' in resp.text:
            raise InvalidScopeError("Invalid scope name in request URL.")
        elif resp.status_code == 401:
            # When API key is not autorized, e.g. an invalid API key.
            raise AuthorizationError("Authorization required. Check if API key and API secret is correct and authorized.")
        elif resp.status_code == 403:
            # When API key has no change permission, e.g. monitoring only API key.
            raise AccessForbiddenError("Secure Workload understood the request but refuses to authorize it(must be Site Admin, Customer Support or a root scope Owner).")
        elif resp.status_code == 502:
            # When API key is invalid or has no change permission, server side may occationaly return 502.
            raise AccessFailed("API key is either invalid or has no change permission.")
        else:
            raise UnknownError("Quarantine operation failed, reason unknown.")

    def is_quarantine_annotation_present(self, ip, action):
        time_to_wait = 150
        interval = 5
        while time_to_wait > 0:
            restclient = RestClient("https://" + self.active_ip, credentials_file="api_credentials.json", verify=False)
            resp = restclient.get('/openapi/v1/assets/cmdb/download/' + self.scope)
            if resp.status_code == 200:
                resp_list = resp.text.splitlines()
                # First line is the header format
                index = self.get_quarantine_index(resp_list[0])
                if index == -1: # Didn't find 'quarantine' in the header
                    return False
                else:
                    for line in resp_list:
                        column_list = line.split(',')
                        if column_list[0] == str(ip) and column_list[index] == 'yes' and action == 'add':
                            return True
                        if column_list[0] == str(ip) and column_list[index] == 'no' and action  == 'delete':
                            return False

            time.sleep(interval)
            time_to_wait -= interval
        return False

    def get_quarantine_index(self, header):
        if header is None:
            return -1
        column_list = header.split(',')
        index = 0;
        for column in column_list:
            if column == 'quarantine':
                return index
            else:
                index += 1
        return -1


    def log_msg(self, filename, prefix, msg):
        if not debug:
            return
        f = open(filename, "a")
        # restrict file size to 5*1024*1024, this message amount should be sufficient for trouble-shoot
        if f.tell() < 5*1024*1024:
            f.write(prefix)
            if msg:
                f.write(msg)
            f.write('\n')
        f.close()

